import time
import torch
import random
import torch.nn as nn
import torch.nn.functional as F
import numpy as np

from transformers import AdamW, get_linear_schedule_with_warmup
from colbert.infra import ColBERTConfig
from colbert.training.rerank_batcher import RerankBatcher

from colbert.utils.amp import MixedPrecisionManager
from colbert.training.lazy_batcher import LazyBatcher, MultiLangBatcher
from colbert.parameters import DEVICE

from colbert.modeling.colbert import ColBERT
from colbert.modeling.reranker.electra import ElectraReranker

from colbert.utils.utils import print_message
from colbert.training.utils import print_progress, manage_checkpoints, find_last_checkpoint, load_checkpoint_misc



def train(config: ColBERTConfig, triples, queries=None, collection=None):
    if config.resume:
        config.checkpoint = config.checkpoint or find_last_checkpoint(config.checkpoint_path_)
    else: 
        config.checkpoint = config.checkpoint or config.model_name

    if config.rank < 1:
        config.help()

    random.seed(12345)
    np.random.seed(12345)
    torch.manual_seed(12345)
    torch.cuda.manual_seed_all(12345)

    assert config.bsize % config.nranks == 0, (config.bsize, config.nranks)
    config.bsize = config.bsize // config.nranks

    print("Using config.bsize =", config.bsize, "(per process) and config.accumsteps =", config.accumsteps)

    if collection is not None:
        if hasattr(collection, 'load_all_collections'):
            collection.load_all_collections()
            
        if config.reranker:
            reader = RerankBatcher(config, triples, queries, collection, (0 if config.rank == -1 else config.rank), config.nranks)
        elif config.multilang:
            reader = MultiLangBatcher(config, triples, queries, collection, (0 if config.rank == -1 else config.rank), config.nranks)
        else:
            reader = LazyBatcher(config, triples, queries, collection, (0 if config.rank == -1 else config.rank), config.nranks)
    else:
        reader = LazyBatcher(config, triples, None, None, (0 if config.rank == -1 else config.rank), config.nranks)
        # raise NotImplementedError()

    if not config.reranker:
        colbert = ColBERT(name=config.checkpoint, colbert_config=config)
    else:
        colbert = ElectraReranker.from_pretrained(config.checkpoint)

    colbert = colbert.to(DEVICE)
    colbert.train()

    colbert = torch.nn.parallel.DistributedDataParallel(colbert, device_ids=[config.rank],
                                                        output_device=config.rank,
                                                        find_unused_parameters=True)

    optimizer = AdamW(filter(lambda p: p.requires_grad, colbert.parameters()), lr=config.lr, eps=1e-8)
    optimizer.zero_grad()

    scheduler = None
    if config.warmup is not None:
        print(f"#> LR will use {config.warmup} warmup steps and linear decay over {config.maxsteps} steps.")
        scheduler = get_linear_schedule_with_warmup(optimizer, num_warmup_steps=config.warmup,
                                                    num_training_steps=config.maxsteps)

    warmup_bert = config.warmup_bert
    if warmup_bert is not None:
        set_bert_grad(colbert, False)

    amp = MixedPrecisionManager(config.amp)
    labels = torch.zeros(config.bsize, dtype=torch.long, device=DEVICE)

    start_time = time.time()
    train_loss = None
    train_loss_mu = 0.999

    start_batch_idx = 0

    if config.resume or config.resume_optimizer:
        start_batch_idx, optimizer_state_dict = load_checkpoint_misc(config)
        optimizer.load_state_dict(optimizer_state_dict)

        if config.resume:
            reader.skip_to_batch(start_batch_idx)

    for batch_idx, BatchSteps in zip(range(start_batch_idx, config.maxsteps), reader):
        if (warmup_bert is not None) and warmup_bert <= batch_idx:
            set_bert_grad(colbert, True)
            warmup_bert = None

        this_batch_loss = 0.0

        for batch in BatchSteps:
            with amp.context():
                if len(batch) == 3:
                    queries, passages, target_scores = batch
                    encoding = [queries, passages]
                else:
                    encoding, target_scores = batch
                    encoding = [encoding.to(DEVICE)]

                scores: torch.Tensor = colbert(*encoding)

                if config.use_ib_negatives:
                    scores, ib_loss = scores

                scores = scores.view(-1, config.nway)

                if len(target_scores) and not config.ignore_scores:
                    target_scores = torch.tensor(target_scores).view(-1, config.nway).to(DEVICE)
                    target_scores = target_scores * config.distillation_alpha
                    target_scores = F.log_softmax(target_scores, dim=-1)

                    log_scores = F.log_softmax(scores, dim=-1)
                    
                    if config.kd_loss == 'KLD':
                        loss = nn.KLDivLoss(reduction='batchmean', log_target=True)(log_scores, target_scores)
                    elif config.kd_loss == 'MSE':
                        loss = nn.MSELoss()(log_scores, target_scores)
                    else:
                        raise ValueError(f"Unsupported kd loss function: {config.kd_loss}")
                else:
                    # if config.rank < 1:
                    #     print(batch)
                    #     print(scores, labels[:scores.size(0)])
                    loss = nn.CrossEntropyLoss()(scores, labels[:scores.size(0)])

                if config.use_ib_negatives: # Eugene: pretty much broken...
                    if config.rank < 1:
                        print('\t\t\t\t', loss.item(), ib_loss.item())

                    loss += ib_loss
                
                if config.multilang and not config.nolangreg:
                    ## all passages might be too much...
                    # lang_scores: torch.Tensor = scores.view(-1, config.bsize // config.accumsteps, config.nway).permute(1, 2, 0).flatten(start_dim=0, end_dim=1)
                    lang_scores: torch.Tensor = log_scores.view(-1, config.bsize // config.accumsteps, config.nway).permute(1, 2, 0).flatten(start_dim=0, end_dim=1)
                    
                    # try only the most relevant passage
                    # most_rel = target_scores.argmax(dim=-1)
                    # lang_scores = scores.index_select(-1, most_rel).diag().view(reader.nlang, -1).T
                    # print(scores)
                    # print(lang_scores)

                    # loss += (lang_scores.max(dim=-1).values - lang_scores.min(dim=-1).values).mean()
                    # lang_loss = (lang_scores - lang_scores.mean(dim=-1).unsqueeze(1)).norm(dim=-1).mean() 
                    # scale = min(1 / (200*(loss.item()**2)), 1.)
                    
                    lang_loss = []
                    for i in range(reader.nlang):
                        for j in range(i, reader.nlang):
                            lang_loss.append(symmetric_divergence(lang_scores[:, i], lang_scores[:, j]))
                    lang_loss = torch.stack(lang_loss).mean()
                    
                    if config.rank < 1:
                        print(f"#>>> loss={loss.item()}+{lang_loss.item()}")
                        
                    loss += lang_loss


                loss = loss / config.accumsteps

            if config.rank < 1:
                if len(target_scores) and not config.ignore_scores:
                    highest_avg, lowest_avg = scores.max(dim=-1).values.mean().item(), scores.min(dim=-1).values.mean().item()
                    print(f"#>>>   {highest_avg:.2f}, {lowest_avg:.2f} \t\t|\t\t {highest_avg-lowest_avg:.2f}")
                else:
                    print_progress(scores)

            amp.backward(loss)

            this_batch_loss += loss.item()

        train_loss = this_batch_loss if train_loss is None else train_loss
        train_loss = train_loss_mu * train_loss + (1 - train_loss_mu) * this_batch_loss

        amp.step(colbert, optimizer, scheduler)

        if config.rank < 1:
            print_message(batch_idx, train_loss)
            manage_checkpoints(config, colbert, optimizer, batch_idx+1, savepath=None)

    if config.rank < 1:
        print_message("#> Done with all triples!")
        ckpt_path = manage_checkpoints(config, colbert, optimizer, batch_idx+1, savepath=None, consumed_all_triples=True)

        return ckpt_path  # TODO: This should validate and return the best checkpoint, not just the last one.


def symmetric_divergence(dist_a, dist_b, alpha=0.5, log_target=True):
    # Jensen–Shannon divergence
    return alpha * nn.KLDivLoss(reduction='batchmean', log_target=log_target)(dist_a, dist_b) + \
           (1-alpha) * nn.KLDivLoss(reduction='batchmean', log_target=log_target)(dist_b, dist_a)


def set_bert_grad(colbert, value):
    try:
        for p in colbert.bert.parameters():
            assert p.requires_grad is (not value)
            p.requires_grad = value
    except AttributeError:
        set_bert_grad(colbert.module, value)



import argparse
import queue
from dataclasses import dataclass
import datetime
from typing import Tuple
"""
Usage: python -m colbert.scripts.shirttt_utils --input <ranking files from each shard> --output <file to write> --topn 50 

"""

@dataclass(order=True)
class PrioritizedItem:
    priority: float  # -1 * score
    item: Tuple[int, list] # index, line from output file

class ResultData:
    def __init__(self, qry_list):
        self.cur = 0
        self.ranked_list = qry_list

def create_item(filename, index):
    with open(filename) as fin:
        for i, line in enumerate(fin):
            split = line.split()
            score = float(split[4])*-1
            qid = split[0]
            pitem = PrioritizedItem(score, (index, split))
            yield qid, pitem

def print_message(*s, condition=True, pad=False):
    s = ' '.join([str(x) for x in s])
    msg = "[{}] {}".format(datetime.datetime.now().strftime("%b %d, %H:%M:%S.%f"), s)

    if condition:
        msg = msg if not pad else f'\n{msg}\n'
        print(msg, flush=True)

def combine_shards(args):
    results = {}
    print_message('Begin merge')
    for i, fn in enumerate(args.input):
        query_list = []
        query_id = None
        for qid, item in create_item(fn, i):
            if qid != query_id:
                if query_list:
                    if not qid in results:
                        results[qid] = [None]*len(args.input)
                    results[qid][i] = ResultData(query_list)
                query_id = qid
                query_list = []
            if args.pershard < 0 or len(query_list) < args.pershard:
                query_list.append(item)
        if query_list:
            if not qid in results:
                results[qid] = [None]*len(args.input)
            results[qid][i] = ResultData(query_list)

    print_message('Read data')

    with open(args.output, 'w') as fout:
        for qid, data in results.items():
            pq = queue.PriorityQueue(len(args.input))

            for q_results in data:
                if q_results.ranked_list and q_results.cur < len(q_results.ranked_list):
                    pq.put(q_results.ranked_list[q_results.cur])
                    q_results.cur += 1
            for i in range(args.topn):
                if not pq.empty():
                    # Get the next document
                    top_ranked = pq.get()
                    doc_info = top_ranked.item[1]
                    doc_info[3] = str(i+1)
                    print(' '.join(doc_info), file=fout)
                    #Replace the document from the correct shard
                    idx = top_ranked.item[0]
                    if data[idx] and data[idx].cur < len(data[idx].ranked_list):
                        pq.put(data[idx].ranked_list[data[idx].cur])
                        data[idx].cur += 1
                else:
                    break

    print_message('All done')

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--input', help='trec files', required=True, nargs='+')
    parser.add_argument('--output', required=True)
    parser.add_argument('--topn', help='number of documents to rank', type=int, required=True)
    parser.add_argument('--pershard', help='number of documents to rank from a shard', type=int, default=-1)

    combine_shards(parser.parse_args())



from pathlib import Path
from typing import Dict, List, Union

import gzip
import pickle
import json
from time import time
import random
import re
from tqdm import tqdm

from colbert.infra import Run, ColBERTConfig
from colbert.utils.utils import print_message, easy_pbar
from colbert.data import Queries, Collection, Examples

from huggingface_hub import hf_hub_download
from huggingface_hub.utils import EntryNotFoundError

import ir_datasets as irds
import ir_measures as irms

import logging

_component_cls_map: Dict[str, Union[Examples, Queries, Collection]] = { 
    'docpairs': Examples, 'queries': Queries, 'docs': Collection 
}


def load_mapping(f: str, is_dummy: bool=False) -> Dict[int, str]:
    return { 
        int(pid): "_".join(tag.split("_")[:-1]) if not is_dummy else tag
        for pid, tag in map(lambda l: l.strip().split("\t"), open(f)) 
    }


def strip_newlines(text: str) -> str:
    text = text.replace("\n", " ")
    text = re.sub("\s+", " ", text)
    return text


class _irds_wrapper:
    def __init__(self, dsid, component='docs'):
        self.dsid = dsid
        self.component = component

        self.id_key = 'doc_id' if component == 'docs' else 'query_id'
    
    @property
    def data(self):
        return getattr(irds.load(self.dsid), self.component)

    def _process(self, d):
        return d.text

    def _get_id(self, d):
        return getattr(d, self.id_key)

    def __getitem__(self, i):
        if isinstance(i, int):
            i = str(i)
        return self._process(self.data.lookup(i))
    
    def __iter__(self):
        return map(self._process, iter(self.data))

    def __len__(self):
        return len(self.data)
    
    def items(self):
        yield from ( (self._get_id(d), self._process(d)) for d in self.data )


class OffsetMapCollection(Collection):
    def __init__(self, config: ColBERTConfig, collection_path: str, offsetmap_path: str=None, build_offset: bool=True):
        self.config = config
        assert collection_path.endswith('.tsv') or collection_path.endswith('.tsv.gz'), "Only support tsv for now"
        self.collection_path = collection_path

        if build_offset:
            self.path, self.mapping = self._build_or_load_offset_map(collection_path, offsetmap_path)
            self.length = len(self.mapping)
        else:
            self.length = self._count_num_lines()
            self.path, self.mapping = None, None
        
        self._reader = None
    
    def _get_opener(self):
        return (gzip.open if self.collection_path.endswith('.gz') else open)

    @property
    def reader(self):
        # avoid the file pointer being pickled
        if self._reader is None or self._reader.closed:
            self._reader = (gzip.open if self.collection_path.endswith('.gz') else open)(self.collection_path, 'rt')
        return self._reader

    def _count_num_lines(self):
        return sum(1 for _ in easy_pbar(self._get_opener()(self.collection_path, 'rt'), "counting length", disabled=False, use_tqdm=True))

    def _load_offset_map(self, offsetmap_path: str, collection_path: str):
        if not offsetmap_path:
            return None
        
        target_path, mapping =  pickle.load(open(offsetmap_path, 'rb'))
        assert Path(target_path).absolute() == Path(collection_path).absolute(), f"Mapping was built for {target_path} not {collection_path}."
        return str(offsetmap_path), mapping

    def _build_or_load_offset_map(self, collection_path: str, offsetmap_path: str=None):
        if offsetmap_path is None:
            index_path = Path(self.config.index_path_)
            index_path.mkdir(parents=True, exist_ok=True)
            offsetmap_path = str(index_path / "offset_map.pkl")

        if Path(offsetmap_path).exists():
            return self._load_offset_map(offsetmap_path, collection_path)

        mapping = []
        
        # explicitly avoid touching self.reader
        with self._get_opener()(self.collection_path, 'rt') as fr:
            for i, _ in enumerate(easy_pbar(desc='building offset map', disabled=False, use_tqdm=True)): # while True
                loc = fr.tell()
                line = fr.readline()
                if line == '':
                    break
                assert int(line.split('\t')[0]) == i
                mapping.append(loc)
        
        print(f"> Writing offset map to {offsetmap_path}")
        with open(offsetmap_path, 'wb') as fw:
            pickle.dump((str(collection_path), mapping), fw)
            fw.flush()
        
        return offsetmap_path, mapping

    def _parse_line(self, line: str):
        pid, passage, *rest = line.strip('\n\r ').split('\t')
        return (rest[0] + ' | ' + passage) if len(rest) >= 1 else passage
        
    def __iter__(self):
        self.reader.seek(0)
        yield from map(
            self._parse_line, 
            easy_pbar(self.reader, desc="reading collection on-the-fly", disabled=Run().config.rank != 0)
            # one process displaying the progress would be enough
        )

    def __getitem__(self, idx: Union[int, List[int]]):
        assert self.mapping is not None, f"Have not built the offset map."

        if isinstance(idx, list):
            return [ self.__getitem__(i) for i in idx ]

        self.reader.seek(self.mapping[int(idx)])
        return self._parse_line(self.reader.readline())

    def __len__(self):
        return self.length
    
    def save(self, new_path):
        raise NotImplementedError(f"Offset map collection is not meant to be stored. ")


def _load_single_collection(c: Union[str, Collection], use_offsetmap: bool=False):
    if isinstance(c, str) and use_offsetmap:
        print_message(f"Loading {c} with offsetmap...")
        p = Path(c)
        if (p.parent / f"{p.name}.offsetmap").exists():
            return OffsetMapCollection(
                collection_path=c, 
                offsetmap_path=str(p.parent / f"{p.name}.offsetmap"),
                config=None
            )
    return Collection.cast(c)


class MutliCollections(Collection):
    def __init__(self, children, mixing='elements', load_now=False, use_offsetmap=False):
        self.children = children
        self.loaded = False
        if load_now:
            self.load_all_collections()

        assert mixing in ['entries', 'elements', 'all']
        self.strategy = mixing
        print_message(f"Using {len(self.children)} collections with {self.strategy} mixing strategy.")

        self.use_offsetmap = use_offsetmap
    

    def load_all_collections(self):
        self.children = [ _load_single_collection(c, self.use_offsetmap) for c in self.children ]
        assert all([ len(self.children[0]) == len(c) for c in self.children ])
        self.loaded = True

    def _choice(self, options):
        # TODO: could be fancier
        return random.choice(options)

    def __getitem__(self, item):
        assert self.loaded

        # TODO: might need some checking
        if self.strategy == 'entries':
            assert isinstance(item, list)
            return self._choice(self.children)[item]

        elif self.strategy == 'elements':
            if not isinstance(item, list):
                return self._choice(self.children)[item]
            return [ self._choice(self.children)[i] for i in item ]
        
        elif self.strategy == 'all':
            selected = [ c[item] for c in self.children ]
            if not isinstance(item, list):
                return selected
            # change to [ [i1l1, i1l2, ...], [i2l1, i2l2, ...] ]
            return list(map(list, zip(*selected))) # transpose

        raise ValueError(f"strategy {self.strategy} not supported.")
        # return self._choice([ c[item] for c in self.children ])

    def __iter__(self):
        if self.strategy == 'elements':
            yield from map(self._choice, zip(*self.children))
        elif self.strategy == 'all':
            yield from map(list, zip(*self.children))

        raise ValueError(f"strategy {self.strategy} not supported.")

    def __len__(self):
        return len(self.children[0])

    def provenance(self):
        return f"Combination({self.strategy})[{'+'.join([ c.provenance() if isinstance(c, Collection) else str(c) for c in self.children ])}]"


def load_irds_or_local(ds: Union[List[str],str], component: str, mixing='elements', use_offsetmap=False) -> Union[Examples, Collection, Examples, str]:
    if isinstance(ds, list):
        if len(ds) > 1:
            # TODO: support queries or even examples
            return MutliCollections([load_irds_or_local(d, component) for d in ds], mixing, use_offsetmap=use_offsetmap)
        ds = ds[0]

    if ds in irds.registry:
        print_message(f"Use IRDS:{ds} for {component}")
        if component == 'qrels':
            return list(irds.load(ds).qrels_iter())
        return _component_cls_map[component](data=_irds_wrapper(ds, component), path=f"irds:{ds}:{component}")
    else:
        print_message(f"Use local file:{ds} for {component}")
        assert Path(ds).exists(), f"File {ds} does not exists."

        if component == 'qrels':
            return list(irms.read_trec_qrels(ds))

        # avoid materialize it in the main process if using raw file
        if component == "docs":
            return _load_single_collection(ds, use_offsetmap=True) if use_offsetmap else ds
        return _component_cls_map[component](path=ds)

def irds_contains(ds: Union[List[str], str], component: str):
    if isinstance(ds, list):
        return [ irds_contains(d, component) for d in ds ]
    return ds in irds.registry and irds.load(ds).has(component)


def _offsetmap_cache_cli(args):
    collection_path: Path = args.collection_path
    offsetmap_path = collection_path.parent / f"{collection_path.name}.offsetmap"
    print_message(collection_path)
    OffsetMapCollection(None, str(collection_path), offsetmap_path, build_offset=True)


def _process_documents(args):
    if not Path(args.corpus).exists():
        try:
            args.corpus = hf_hub_download(*str(args.corpus).split(":"), repo_type="dataset")
        except EntryNotFoundError:
            raise FileNotFoundError(f"`{args.corpus}` does not exist on disk nor Huggingface Hub")

    opener = gzip.open if args.corpus.endswith('.gz') else open
    num_docs = sum(1 for _ in tqdm(opener(args.corpus, 'rt'), desc='counting'))
    with opener(args.corpus, 'rt') as f:
        for i, line in tqdm(enumerate(f), total=num_docs):
            try:
                temp = json.loads(line)
            except json.decoder.JSONDecodeError:
                print(f"json decode error on line #{i} -- `{line}`")
                continue
            doc_id = temp[args.docid]
            title = strip_newlines(temp[args.title]) if args.title in temp else ""
            text = strip_newlines(temp[args.body])
            doc_text = title + " " + text if title else text
            if args.lower:
                doc_text = doc_text.lower()
            encoded_tokens = args.tokenizer.encode(doc_text)[1:-1]
            s, e, idx = 0, 0, 0
            while s < len(encoded_tokens):
                e = s + args.length
                if e >= len(encoded_tokens):
                    e = len(encoded_tokens)
                p = args.tokenizer.decode(encoded_tokens[s:e])
                pass_id = f"{doc_id}_{idx}"
                s = s + args.length - args.stride
                yield p, pass_id
                idx += 1


def _create_passage_collection_cli(args):
    from transformers import AutoTokenizer
    logging.getLogger("transformers.tokenization_utils_base").setLevel(logging.ERROR) 

    root: Path = args.root
    if not root.exists():
        root.mkdir(parents=True)
    args.tokenizer = AutoTokenizer.from_pretrained(args.tokenizer)
    pass_file= root / "collection_passages.tsv"
    map_file = root / "mapping.tsv"
    
    with open(pass_file, "w") as f, open(map_file, "w") as g:
        for idx,(line1,line2) in enumerate(_process_documents(args)):
            f.write(f"{idx}\t{line1}\n")
            g.write(f"{idx}\t{line2}\n")


def _merge_teacher_scores(args):
    assert len(args.input_files) > 1, "We can only merge >1 files"
    
    combined = {}
    
    for fn in tqdm(args.input_files, desc="reading file...", dynamic_ncols=True):
        with open(fn) as fr:
            for triple_list in tqdm(map(json.loads, fr), dynamic_ncols=True):
                qid = triple_list[0]
                assert isinstance(qid, str)
                if qid not in combined:
                    combined[qid] = []
                combined[qid] += triple_list[1:]
    
    with open(args.output_file, "w") as fw:
        for qid, triple_list in tqdm(combined.items(), total=len(combined), desc='writing...'):
            fw.write(json.dumps([qid] + triple_list) + "\n")


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser("Collection helper cli")
    subparser = parser.add_subparsers()

    offsetmap_parser = subparser.add_parser("build_offsetmap")
    offsetmap_parser.add_argument('collection_path', type=Path)
    offsetmap_parser.set_defaults(func=_offsetmap_cache_cli)

    passaging_parser = subparser.add_parser("create_passage_collection")
    passaging_parser.add_argument("--root", type=Path, required=True)
    passaging_parser.add_argument("--corpus", type=Path, required=True)
    passaging_parser.add_argument("--length", type=int, default=180)
    passaging_parser.add_argument("--stride", type=int, default=90)
    passaging_parser.add_argument("--docid", type=str, default="id")
    passaging_parser.add_argument("--title", type=str, default="title")
    passaging_parser.add_argument("--body", type=str, default="text")
    passaging_parser.add_argument("--lower", action="store_true", default=False) 
    passaging_parser.add_argument('--tokenizer', type=str, default="xlm-roberta-large")
    passaging_parser.set_defaults(func=_create_passage_collection_cli)

    teacher_merging_parser = subparser.add_parser("merge_teacher_scores")
    teacher_merging_parser.add_argument("--input_files", nargs='+', type=Path, required=True)
    teacher_merging_parser.add_argument("--output_file", type=Path, required=True)
    teacher_merging_parser.set_defaults(func=_merge_teacher_scores)

    args = parser.parse_args()
    args.func(args)



import argparse
from pathlib import Path
from typing import Dict, List, Tuple
from colbert.infra import Run, RunConfig, ColBERTConfig
from colbert import Searcher

from tqdm.auto import tqdm

import ir_measures as irms
from colbert.scripts.collection_utils import load_mapping, load_irds_or_local, irds_contains

def maxp(passage_triples: List[Tuple[str, int, float]], mapping: Dict[str, str]) -> Dict[str, float]:
    doc_score = {}
    for pid, _, score in passage_triples:
        doc_id = mapping[pid]
        if doc_id not in doc_score or doc_score[doc_id] < score:
            doc_score[doc_id] = score
    return doc_score

def write_trec_run(doc_scores: Dict[str, Dict[str, float]], fn: str, run_id: str = 'PLAID', topk=1000):
    with open(fn, 'w') as fw:
        for qid, d in doc_scores.items():
            for i, (did, s) in enumerate(sorted(d.items(), key=lambda x:x[1], reverse=True)[:topk]):
                fw.write(f"{qid} 0 {did} {i+1} {s} {run_id}\n")

def main(args):
    output_fn = args.output or f"./{args.run_id}.{args.search_depth}.trec"
    if args.check_exists and Path(output_fn).exists():
        print(f"Found {output_fn} -- skip")
        return

    queries = load_irds_or_local(args.queries, component='queries', use_offsetmap=False)
    mapping = load_mapping(args.passage_mapping, is_dummy=not args.maxp)

    if args.run_evel:
        metrics = [ irms.parse_measure(m) for m in args.metrics ]
        qrels = load_irds_or_local(args.qrels, component='qrels')

    with Run().context(RunConfig(nranks=args.n_gpus, root=args.root, experiment=args.experiment, index_root=args.index_root)):
        checkpoint = None
        if args.checkpoint_root is not None:
            checkpoint = args.checkpoint_root + "/" + ColBERTConfig.load_from_index( args.index_root + "/" + args.index_name ).checkpoint
        searcher = Searcher(index=args.index_name, checkpoint=checkpoint)
        searcher.config.configure(only_approx=args.only_approx, ignore_unrecognized=False)
        
        rankings = searcher.search_all(queries, k=args.search_depth)

    doc_ranking = { 
        str(qid): maxp(r, mapping) for qid, r in tqdm(rankings.items(), desc='MaxP') 
    }

    if not args.no_save:
        print(f"Saving results to {output_fn}")
        write_trec_run(doc_ranking, output_fn, args.run_id)

    if args.run_evel:
        print( irms.calc_aggregate(metrics, qrels, doc_ranking) )


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--index_name', help="name of the index", type=str)
    parser.add_argument('--passage_mapping', help="mapping tsv file from passage to documents", type=str)
    parser.add_argument('--query_file', '--query', '--queries', dest='queries', help="query tsv file or irds", type=str, required=True)
    parser.add_argument('--search_depth', help="k; depth of PLAID search", type=int, default=2500)
    parser.add_argument('--centroid_score_threshold', help="threshold for centroid cell filter", type=float, default=0.4)

    parser.add_argument('--maxp', action='store_true', default=False)
    parser.add_argument('--checkpoint_root', type=str, default=None)

    parser.add_argument('--metrics', help="evalutation metrics used after search, should be parsable by ir_measures", 
                        type=str, nargs='*', default=[])
    parser.add_argument('--qrel', '--qrels', dest='qrels', help="path to qrels file", type=str, default=None)

    parser.add_argument('--run_id', help="name of this run, will be used for the filename of the trec file", type=str, default=None)
    parser.add_argument('--output', help="output directory of the trec file", type=str, default=None)

    parser.add_argument('--index_root', help="root directory of the indexes", type=str, default=None)
    parser.add_argument('--experiment', help="name of the experiment for locating the index", type=str, default="default")
    parser.add_argument('--root', type=str, default=None)
    parser.add_argument('--n_gpus', help="number of gpus to use, default all visible ones in CUDA_VISIBLE_DEVICES", type=int, default=None)

    parser.add_argument('--only_approx', action='store_true', default=False)
    parser.add_argument('--no_save', action='store_true', default=False)

    parser.add_argument('--check_exists', action='store_true', default=False)

    parser.set_defaults(run_evel=False)

    args = parser.parse_args()

    if args.n_gpus is None:
        import torch
        args.n_gpus = torch.cuda.device_count()
    
    if len(args.metrics) > 0:
        if args.qrels is None and irds_contains(args.queries, 'qrels'):
            args.qrels = args.queries

        assert args.qrels and (Path(args.qrels).exists() or irds_contains(args.qrels, 'qrels')), args.qrels
        args.run_evel = True

    assert Path(args.queries).exists() or irds_contains(args.queries, 'queries'), args.queries

    if args.run_id is None:
        args.run_id = f"{args.experiment}.{args.index_name}.{Path(args.queries).name}"
        args.run_id = args.run_id.replace('/', '.')

    main(args)


import os
import tempfile
from typing import Dict, Iterable
import tqdm
import time
import ujson
import torch
import pickle
import random
from collections import defaultdict
import more_itertools
from pathlib import Path

try:
    import faiss
except ImportError as e:
    print("WARNING: faiss must be imported for indexing")

import numpy as np
import torch.multiprocessing as mp
from colbert.infra.config.config import ColBERTConfig

import colbert.utils.distributed as distributed

from colbert.infra.run import Run
from colbert.infra.launcher import print_memory_stats
from colbert.modeling.checkpoint import Checkpoint
from colbert.data.collection import Collection

from colbert.indexing.collection_encoder import CollectionEncoder
from colbert.indexing.index_saver import IndexSaver
from colbert.indexing.loaders import load_doclens
from colbert.indexing.utils import optimize_ivf
from colbert.utils.utils import flatten, print_message

from colbert.indexing.codecs.residual import ResidualCodec

# Eugene: deprecated
# def encode(config, collection, shared_lists, shared_queues):
#     encoder = CollectionIndexer(config=config, collection=collection)
#     encoder.run(shared_lists)

def reuse_prepare(config, collection, shared_lists, shared_queues):
    encoder = CollectionIndexer(config=config, collection=collection, no_checkpoint=True)
    encoder.steps(shared_lists, -1)

def sample(config, collection, shared_lists, shared_queues):
    encoder = CollectionIndexer(config=config, collection=collection)
    encoder.steps(shared_lists, 0)

def kmeans(config, collection, shared_lists, shared_queues):
    encoder = CollectionIndexer(config=config, collection=collection, no_checkpoint=True)
    encoder.steps(shared_lists, 1)

def encode(config, collection, shared_lists, shared_queues):
    encoder = CollectionIndexer(config=config, collection=collection)
    encoder.steps(shared_lists, 2)

def finalize(config, collection, shared_lists, shared_queues):
    encoder = CollectionIndexer(config=config, collection=collection)
    encoder.steps(shared_lists, 3)


class CollectionIndexer():
    '''
    Given a collection and config, encode collection into index and
    stores the index on the disk in chunks.
    '''
    def __init__(self, config: ColBERTConfig, collection, no_checkpoint=False):
        self.config = config
        self.rank, self.nranks = self.config.rank, self.config.nranks

        self.use_gpu = self.config.total_visible_gpus > 0

        if self.config.rank == 0:
            self.config.help()

        self.collection = Collection.cast(collection)

        if not no_checkpoint:
            self.checkpoint = Checkpoint(self.config.checkpoint, colbert_config=self.config)
            if self.use_gpu:
                self.checkpoint = self.checkpoint.cuda()

            self.encoder = CollectionEncoder(config, self.checkpoint)
        
        self.saver = IndexSaver(config)

        print_memory_stats(f'RANK:{self.rank}')

    def run(self, shared_lists):
        # FIXME Eugene: 
        # For some reason running `self.train` on rank=0 is either not sufficient to block other
        # child processes to use GPU or preventing FAISS from allocating memory on GPUs previously used by other child processes.
        # So to get around this, the pipeline needs to be separated into three parts -- setup(multiprocess), 
        # train(initiating from the main process), and the rest(multiprocessing). 
        # This would be controlled by the indexer. 
        # This function would be deprecated.

        with torch.inference_mode():
            self.setup() # Computes and saves plan for whole collection
            distributed.barrier(self.rank)
            print_memory_stats(f'RANK:{self.rank}')

            if not self.config.resume or not self.saver.try_load_codec():
                self.train(shared_lists) # Trains centroids from selected passages
            distributed.barrier(self.rank)
            print_memory_stats(f'RANK:{self.rank}')

            self.index() # Encodes and saves all tokens into residuals
            distributed.barrier(self.rank)
            print_memory_stats(f'RANK:{self.rank}')

            self.finalize() # Builds metadata and centroid to passage mapping
            distributed.barrier(self.rank)
            print_memory_stats(f'RANK:{self.rank}')

    def steps(self, shared_lists, step: int):
        with torch.inference_mode():
            if step == -1:
                # meant to be in the main process
                self.setup_reuse()
            elif step == 0:
                self.setup()
                distributed.barrier(self.rank)
                print_memory_stats(f'RANK:{self.rank}')
            elif step == 1:
                # should use the main process for entering this block
                if not self.saver.try_load_codec():
                    assert self._try_load_plan()
                    self.train(shared_lists)
            elif step == 2:
                assert self._try_load_plan()
                assert self.saver.try_load_codec()
                self.index()
                distributed.barrier(self.rank)
                print_memory_stats(f'RANK:{self.rank}')
            
            elif step == 3:
                # should use the main process for entering this block
                assert self._try_load_plan()

                self.finalize()
                
                print_memory_stats(f'RANK:{self.rank}')
            else:
                Run().print_main(f"Step {step} does not exist -- ignoring")

    def setup(self):
        '''
        Calculates and saves plan.json for the whole collection.
        
        plan.json { config, num_chunks, num_partitions, num_embeddings_est, avg_doclen_est}
        num_partitions is the number of centroids to be generated.
        '''
        if self.config.resume:
            if self._try_load_plan():
                Run().print_main(f"#> Loaded plan from {self.plan_path}:")
                Run().print_main(f"#> num_chunks = {self.num_chunks}")
                Run().print_main(f"#> num_partitions = {self.num_partitions}")
                Run().print_main(f"#> num_embeddings_est = {self.num_embeddings_est}")
                Run().print_main(f"#> avg_doclen_est = {self.avg_doclen_est}")
                return

        self.num_chunks = int(np.ceil(len(self.collection) / self.collection.get_chunksize()))

        # Saves sampled passages and embeddings for training k-means centroids later 
        sampled_pids = self._sample_pids()
        avg_doclen_est = self._sample_embeddings(sampled_pids)

        # Select the number of partitions
        num_passages = len(self.collection)
        self.num_embeddings_est = num_passages * avg_doclen_est

        self.num_partitions = int(2 ** np.floor(np.log2(16 * np.sqrt(self.num_embeddings_est))))
        if self.config.max_num_partitions > 0:
            self.num_partitions = min(self.num_partitions, self.config.max_num_partitions)

        Run().print_main(f'Creaing {self.num_partitions:,} partitions.')
        Run().print_main(f'*Estimated* {int(self.num_embeddings_est):,} embeddings.')

        self._save_plan()

    def setup_reuse(self):
        if not self._try_load_plan(self.config.reuse_centroids_from):
            raise FileNotFoundError(f"Cannot find plans in index `{self.config.reuse_centroids_from}`")
        
        Run().print_main(f"#> Loaded existing plan from {self.config.reuse_centroids_from}:")
        Run().print_main(f"#> num_partitions = {self.num_partitions}")
        Run().print_main(f"#> avg_doclen_est = {self.avg_doclen_est} <-- not accurate but will not recalculate")
        Run().print_main(f"#> replacing other information with the current collection info")

        # replace collection information here
        self.num_chunks = int(np.ceil(len(self.collection) / self.collection.get_chunksize()))

        num_passages = len(self.collection)
        self.num_embeddings_est = num_passages * self.avg_doclen_est
        self._save_plan()

        # and soft link the rest
        for fn in ['centroids', 'avg_residual', 'buckets']:
            os.symlink(os.path.join(self.config.reuse_centroids_from, f"{fn}.pt"), os.path.join(self.config.index_path_, f"{fn}.pt"))


    def _sample_pids(self):
        num_passages = len(self.collection)

        # Simple alternative: < 100k: 100%, < 1M: 15%, < 10M: 7%, < 100M: 3%, > 100M: 1%
        # Keep in mind that, say, 15% still means at least 100k.
        # So the formula is max(100% * min(total, 100k), 15% * min(total, 1M), ...)
        # Then we subsample the vectors to 100 * num_partitions

        typical_doclen = 120  # let's keep sampling independent of the actual doc_maxlen
        sampled_pids = 16 * np.sqrt(typical_doclen * num_passages)
        # sampled_pids = int(2 ** np.floor(np.log2(1 + sampled_pids)))
        sampled_pids = min(1 + int(sampled_pids), num_passages)
        if self.config.max_sampled_pid > 0:
            sampled_pids = min(sampled_pids, self.config.max_sampled_pid)

        sampled_pids = random.sample(range(num_passages), sampled_pids)
        Run().print_main(f"# of sampled PIDs = {len(sampled_pids)} \t sampled_pids[:3] = {sampled_pids[:3]}")

        return list(sorted(set(sampled_pids)))

    def _sample_embeddings(self, sampled_pids):
        # local_pids = self.collection.enumerate(rank=self.rank)
        # local_sample = [passage for pid, passage in local_pids if pid in sampled_pids]

        # Note from Eugene: how sample is distributed across GPUs shouldn't matter -- as long as each example has a home is fine
        # local_sample = [passage for pid, passage in self.collection.enumerate(rank=self.rank) if pid in sampled_pids]

        # This uses random access from the Collection class -- should be much faster than iterating through the collection
        local_sample = [ self.collection[pid] for pid in sampled_pids[self.rank::self.nranks] ]

        local_sample_embs, doclens = self.encoder.encode_passages(local_sample)

        if torch.cuda.is_available():
            if torch.distributed.is_initialized():
                self.num_sample_embs = torch.tensor([local_sample_embs.size(0)]).cuda()
                torch.distributed.all_reduce(self.num_sample_embs)

                avg_doclen_est = sum(doclens) / len(doclens) if doclens else 0
                avg_doclen_est = torch.tensor([avg_doclen_est]).cuda()
                torch.distributed.all_reduce(avg_doclen_est)

                nonzero_ranks = torch.tensor([float(len(local_sample) > 0)]).cuda()
                torch.distributed.all_reduce(nonzero_ranks)
            else:
                self.num_sample_embs = torch.tensor([local_sample_embs.size(0)]).cuda()

                avg_doclen_est = sum(doclens) / len(doclens) if doclens else 0
                avg_doclen_est = torch.tensor([avg_doclen_est]).cuda()

                nonzero_ranks = torch.tensor([float(len(local_sample) > 0)]).cuda()
        else:
            if torch.distributed.is_initialized():
                self.num_sample_embs = torch.tensor([local_sample_embs.size(0)]).cpu()
                torch.distributed.all_reduce(self.num_sample_embs)

                avg_doclen_est = sum(doclens) / len(doclens) if doclens else 0
                avg_doclen_est = torch.tensor([avg_doclen_est]).cpu()
                torch.distributed.all_reduce(avg_doclen_est)

                nonzero_ranks = torch.tensor([float(len(local_sample) > 0)]).cpu()
                torch.distributed.all_reduce(nonzero_ranks)
            else:
                self.num_sample_embs = torch.tensor([local_sample_embs.size(0)]).cpu()

                avg_doclen_est = sum(doclens) / len(doclens) if doclens else 0
                avg_doclen_est = torch.tensor([avg_doclen_est]).cpu()

                nonzero_ranks = torch.tensor([float(len(local_sample) > 0)]).cpu()

        avg_doclen_est = avg_doclen_est.item() / nonzero_ranks.item()
        self.avg_doclen_est = avg_doclen_est

        Run().print(f'avg_doclen_est = {avg_doclen_est} \t len(local_sample) = {len(local_sample):,}')

        torch.save(local_sample_embs.half(), os.path.join(self.config.index_path_, f'sample.{self.rank}.pt'))

        return avg_doclen_est

    def _try_load_plan(self, path=None):
        config = self.config
        self.plan_path = os.path.join(config.index_path_, 'plan.json')
        path = os.path.join(path, 'plan.json') if path is not None else self.plan_path

        if os.path.exists(path):
            with open(path, 'r') as f:
                try:
                    plan = ujson.load(f)
                except Exception as e:
                    return False
                if not ('num_chunks' in plan and
                        'num_partitions' in plan and
                        'num_embeddings_est' in plan and
                        'avg_doclen_est' in plan):
                    return False

                # TODO: Verify config matches
                self.num_chunks = plan['num_chunks']
                self.num_partitions = plan['num_partitions']
                self.num_embeddings_est = plan['num_embeddings_est']
                self.avg_doclen_est = plan['avg_doclen_est']

                if 'num_sample_embs' in plan:
                    self.num_sample_embs = plan['num_sample_embs']

            return True
        else:
            return False

    def _save_plan(self):
        if self.rank < 1:
            config = self.config
            self.plan_path = os.path.join(config.index_path_, 'plan.json')
            Run().print("#> Saving the indexing plan to", self.plan_path, "..")

            with open(self.plan_path, 'w') as f:
                d = {'config': config.export()}
                d['num_chunks'] = self.num_chunks
                d['num_partitions'] = self.num_partitions
                d['num_embeddings_est'] = self.num_embeddings_est
                d['avg_doclen_est'] = self.avg_doclen_est

                if hasattr(self, 'num_sample_embs'):
                    try:
                        d['num_sample_embs'] = int(self.num_sample_embs.detach().cpu())
                    except:
                        d['num_sample_embs'] = int(self.num_sample_embs)

                f.write(ujson.dumps(d, indent=4) + '\n')


    def train(self, shared_lists):
        if self.rank > 0:
            return

        sample, heldout = self._concatenate_and_split_sample()

        centroids = self._train_kmeans(sample, shared_lists)

        print_memory_stats(f'RANK:{self.rank}')
        del sample

        bucket_cutoffs, bucket_weights, avg_residual = self._compute_avg_residual(centroids, heldout)

        print_message(f'avg_residual = {avg_residual}')

        # Compute and save codec into avg_residual.pt, buckets.pt and centroids.pt
        codec = ResidualCodec(config=self.config, centroids=centroids, avg_residual=avg_residual,
                              bucket_cutoffs=bucket_cutoffs, bucket_weights=bucket_weights)
        self.saver.save_codec(codec)

    def _concatenate_and_split_sample(self):
        print_memory_stats(f'***1*** \t RANK:{self.rank}')

        # TODO: Allocate a float16 array. Load the samples from disk, copy to array.
        # if not hasattr(self, 'num_sample_embs'):
        #     self.num_sample_embs = sum([ 
        #         torch.load( os.path.join(self.config.index_path_, f'sample.{r}.pt') ).size(0) for r in range(self.config.nranks) 
        #     ])

        # forced it to float32 when storing already
        sample = torch.empty(self.num_sample_embs, self.config.dim, dtype=torch.float32)
        
        # load with shuffle to avoid materialize them twice
        shuffle_order = torch.randperm(sample.size(0))

        offset = 0
        print_message(f'Loading {self.nranks} sample embedding files, expecting {self.num_sample_embs} samples')
        for r in range(self.nranks):
            sub_sample_path = os.path.join(self.config.index_path_, f'sample.{r}.pt')
            sub_sample: torch.Tensor = torch.load(sub_sample_path)

            # os.remove(sub_sample_path)

            endpos = offset + sub_sample.size(0)
            # sample[offset:endpos] = sub_sample
            sample[ shuffle_order[offset:endpos] ] = sub_sample.float()
            offset = endpos

        assert endpos == sample.size(0), (endpos, sample.size())

        print_memory_stats(f'***2*** \t RANK:{self.rank}')

        # Shuffle and split out a 5% "heldout" sub-sample [up to 50k elements]
        # sample = sample[torch.randperm(sample.size(0))]

        print_memory_stats(f'***3*** \t RANK:{self.rank}')

        heldout_fraction = 0.05
        heldout_size = int(min(heldout_fraction * sample.size(0), 50_000))
        sample, sample_heldout = sample.split([sample.size(0) - heldout_size, heldout_size], dim=0)

        print_memory_stats(f'***4*** \t RANK:{self.rank}')

        return sample, sample_heldout

    def _train_kmeans(self, sample, shared_lists):
        if self.use_gpu:
            torch.cuda.empty_cache()

        do_fork_for_faiss = False  # set to True to free faiss GPU-0 memory at the cost of one more copy of `sample`.

        args_ = [self.config.dim, self.num_partitions, self.config.kmeans_niters]

        if do_fork_for_faiss:
            # For this to work reliably, write the sample to disk. Pickle may not handle >4GB of data.
            # Delete the sample file after work is done.

            # shared_lists[0][0] = sample
            shared_lists[0].append( sample )
            return_value_queue = mp.Queue()

            args_ = args_ + [shared_lists, return_value_queue]
            proc = mp.Process(target=compute_faiss_kmeans, args=args_)

            proc.start()
            centroids = return_value_queue.get()
            proc.join()

        else:
            args_ = args_ + [[[sample]]]
            centroids = compute_faiss_kmeans(*args_)

        centroids = torch.nn.functional.normalize(centroids, dim=-1)
        if self.use_gpu:
            centroids = centroids.half()
        else:
            centroids = centroids.float()

        return centroids

    def _compute_avg_residual(self, centroids, heldout):
        compressor = ResidualCodec(config=self.config, centroids=centroids, avg_residual=None)

        heldout_reconstruct = compressor.compress_into_codes(heldout, out_device='cuda' if self.use_gpu else 'cpu')
        heldout_reconstruct = compressor.lookup_centroids(heldout_reconstruct, out_device='cuda' if self.use_gpu else 'cpu')
        if self.use_gpu:
            heldout_avg_residual = heldout.cuda() - heldout_reconstruct
        else:
            heldout_avg_residual = heldout - heldout_reconstruct

        avg_residual = torch.abs(heldout_avg_residual).mean(dim=0).cpu()
        print([round(x, 3) for x in avg_residual.squeeze().tolist()])

        num_options = 2 ** self.config.nbits
        quantiles = torch.arange(0, num_options, device=heldout_avg_residual.device) * (1 / num_options)
        bucket_cutoffs_quantiles, bucket_weights_quantiles = quantiles[1:], quantiles + (0.5 / num_options)

        bucket_cutoffs = heldout_avg_residual.float().quantile(bucket_cutoffs_quantiles)
        bucket_weights = heldout_avg_residual.float().quantile(bucket_weights_quantiles)

        print_message(
            f"#> Got bucket_cutoffs_quantiles = {bucket_cutoffs_quantiles} and bucket_weights_quantiles = {bucket_weights_quantiles}")
        print_message(f"#> Got bucket_cutoffs = {bucket_cutoffs} and bucket_weights = {bucket_weights}")

        return bucket_cutoffs, bucket_weights, avg_residual.mean()

        # EVENTAULLY: Compare the above with non-heldout sample. If too different, we can do better!
        # sample = sample[subsample_idxs]
        # sample_reconstruct = get_centroids_for(centroids, sample)
        # sample_avg_residual = (sample - sample_reconstruct).mean(dim=0)

    def index(self):
        '''
        Encode embeddings for all passages in collection.
        Each embedding is converted to code (centroid id) and residual.
        Embeddings stored according to passage order in contiguous chunks of memory.

        Saved data files described below:
            {CHUNK#}.codes.pt:      centroid id for each embedding in chunk
            {CHUNK#}.residuals.pt:  16-bits residual for each embedding in chunk
            doclens.{CHUNK#}.pt:    number of embeddings within each passage in chunk
        '''
        with self.saver.thread():
            batches = self.collection.enumerate_batches(rank=self.rank)
            for chunk_idx, offset, passages in tqdm.tqdm(batches, disable=self.rank > 0):
                if self.config.resume and self.saver.check_chunk_exists(chunk_idx):
                    Run().print_main(f"#> Found chunk {chunk_idx} in the index already, skipping encoding...")
                    continue
                # Encode passages into embeddings with the checkpoint model
                embs, doclens = self.encoder.encode_passages(passages) 
                if self.use_gpu:
                    assert embs.dtype == torch.float16
                else:
                    assert embs.dtype == torch.float32
                    embs = embs.half()

                Run().print_main(f"#> Saving chunk {chunk_idx}: \t {len(passages):,} passages "
                                 f"and {embs.size(0):,} embeddings. From #{offset:,} onward.")

                self.saver.save_chunk(chunk_idx, offset, embs, doclens) # offset = first passage index in chunk
                del embs, doclens

    def finalize(self):
        '''
        Aggregates and stores metadata for each chunk and the whole index
        Builds and saves inverse mapping from centroids to passage IDs

        Saved data files described below:
            {CHUNK#}.metadata.json: [ passage_offset, num_passages, num_embeddings, embedding_offset ]
            metadata.json: [ num_chunks, num_partitions, num_embeddings, avg_doclen ]
            inv.pid.pt: [ ivf, ivf_lengths ]
                ivf is an array of passage IDs for centroids 0, 1, ...
                ivf_length contains the number of passage IDs for each centroid
        '''
        if self.rank > 0:
            return

        self._check_all_files_are_saved()
        self._collect_embedding_id_offset()

        if not os.path.exists(os.path.join(self.config.index_path_, 'ivf.pid.pt')):
            self._build_ivf()
        else:
            Run().print_main("Found ivf file -- skip building")
        self._update_metadata()

    def _check_all_files_are_saved(self):
        Run().print_main(f"#> Checking all files were saved, expecting {self.num_chunks}...")
        success = True
        for chunk_idx in range(self.num_chunks):
            if not self.saver.check_chunk_exists(chunk_idx):
                success = False
                Run().print_main(f"#> ERROR: Could not find chunk {chunk_idx}!")
                #TODO: Fail here?
        if success:
            Run().print_main("Found all files!")

    def _collect_embedding_id_offset(self):
        passage_offset = 0
        embedding_offset = 0

        self.embedding_offsets = []

        for chunk_idx in range(self.num_chunks):
            metadata_path = os.path.join(self.config.index_path_, f'{chunk_idx}.metadata.json')
            Run().print_main(f"#> Read file: `{metadata_path}`")

            with open(metadata_path) as f:
                chunk_metadata = ujson.load(f)

                chunk_metadata['embedding_offset'] = embedding_offset
                self.embedding_offsets.append(embedding_offset)

                assert chunk_metadata['passage_offset'] == passage_offset, (chunk_idx, passage_offset, chunk_metadata)

                passage_offset += chunk_metadata['num_passages']
                embedding_offset += chunk_metadata['num_embeddings']

            with open(metadata_path, 'w') as f:
                f.write(ujson.dumps(chunk_metadata, indent=4) + '\n')

        self.num_embeddings = embedding_offset
        assert len(self.embedding_offsets) == self.num_chunks

    def _build_ivf(self):
        # Maybe we should several small IVFs? Every 250M embeddings, so that's every 1 GB.
        # It would save *memory* here and *disk space* regarding the int64.
        # But we'd have to decide how many IVFs to use during retrieval: many (loop) or one?
        # A loop seems nice if we can find a size that's large enough for speed yet small enough to fit on GPU!
        # Then it would help nicely for batching later: 1GB.

        if self.config.use_lagacy_build_ivf: 
            return self._legacy_build_ivf()

        Run().print_main("#> Building IVF...")

        #codes = torch.zeros(self.num_embeddings,).long()
        #print_memory_stats(f'RANK:{self.rank}')

        all_doclens = load_doclens(self.config.index_path_, flatten=False)
        pid_offsets = torch.cumsum(torch.tensor([0] + [ 
            len(doc) for doc in all_doclens 
        ]), dim=0 )[:-1]

        from multiprocessing import Pool
        from functools import partial
        
        working_dir = tempfile.mkdtemp() if self.config.ivf_use_tempdir else self.config.index_path_

        work_ = partial(build_local_ivf, self.config.index_path_, working_dir, self.num_partitions)
        
        Run().print_main(f"#> Building local IVFs with {self.config.ivf_num_processes} processes...")

        with Pool(self.config.ivf_num_processes) as pool:
            list(pool.map(work_, [
                (pid_offsets[chunk_idx], all_doclens[chunk_idx], chunk_idx)
                for chunk_idx in range(self.num_chunks)
            ]))
        
        # local_cluster2pids: Iterable[Dict[int, set]] = peekable(map(
        #     lambda chunk_idx: torch.load(os.path.join(working_dir, f"localivf.{chunk_idx}.pt")),
        #     range(self.num_chunks)
        # ))
        
        # Run().print_main(f"#> Combining local IVFs...")
        # cluster2pid = [ [] for _ in range(self.num_partitions) ]
        # for local in tqdm.tqdm(local_cluster2pids, total=self.num_chunks):
        #     # for i in range(self.num_partitions):
        #     for idx, pids in local.items():
        #         # cluster2pid[i] += list(local[i])
        #         cluster2pid[idx] += list(pids)

        merging_way = self.config.ivf_merging_ways
        expected_merging_it = int(np.ceil(np.log(self.num_chunks)/np.log(merging_way)))
        Run().print_main(f"{merging_way}-way merging, expecting {expected_merging_it} iterations")

        candidate_fns = _balance_file_size(Path(working_dir).glob("localivf.0.*.pkl"))
        merging_it = 1
        while len(candidate_fns) > 1:
            n_merging = int(np.ceil(len(candidate_fns)/merging_way))
            Run().print_main(f"Merging IVF Iteration #{merging_it}: {len(candidate_fns)} files -- doing {n_merging} merging")

            work_ = partial(merge_ivf, working_dir, merging_it)
            with Pool(min(self.config.ivf_num_processes, n_merging)) as pool:
                list(pool.map(work_, enumerate(more_itertools.batched(candidate_fns, merging_way))))

            candidate_fns = _balance_file_size(Path(working_dir).glob(f"localivf.{merging_it}.*.pkl"))
            merging_it += 1

        # read final one
        Run().print_main("Read final IVF file")
        cluster2pid = [ [] for _ in range(self.num_partitions) ]
        merged_fn = candidate_fns[0]
        for idx, pids in _load_file(merged_fn).items():
            cluster2pid[idx] = list(pids)


        Run().print_main(f"#> Creating ivf tensor...")
        ivf_lengths = torch.tensor([ len(c) for c in cluster2pid ])
        ivf = torch.zeros(ivf_lengths.sum(), dtype=torch.int32)
        offset = 0
        for pids in cluster2pid:
            ivf[offset: offset+len(pids)] = torch.tensor(sorted(list(pids)))
            offset += len(pids)

        optimized_ivf_path = os.path.join(self.config.index_path_, 'ivf.pid.pt')
        torch.save((ivf, ivf_lengths), optimized_ivf_path)
        print_message(f"#> Saved optimized IVF to {optimized_ivf_path}")

        # clean up 
        for fn in tqdm.tqdm(list(Path(working_dir).glob("localivf.*.pkl")), desc='clean up local ivf files'):
            fn.unlink()

    def _legacy_build_ivf(self):
        Run().print_main("#> Building IVF with lagacy build ivf code...")
        
        codes = torch.zeros(self.num_embeddings,).long()
        print_memory_stats(f'RANK:{self.rank}')

        Run().print_main("#> Loading codes...")

        for chunk_idx in tqdm.tqdm(range(self.num_chunks)):
            offset = self.embedding_offsets[chunk_idx]
            chunk_codes = ResidualCodec.Embeddings.load_codes(self.config.index_path_, chunk_idx)

            codes[offset:offset+chunk_codes.size(0)] = chunk_codes

        assert offset+chunk_codes.size(0) == codes.size(0), (offset, chunk_codes.size(0), codes.size())

        Run().print_main(f"Sorting codes...")

        print_memory_stats(f'RANK:{self.rank}')

        codes = codes.sort()
        ivf, values = codes.indices, codes.values

        print_memory_stats(f'RANK:{self.rank}')

        Run().print_main(f"Getting unique codes...")

        ivf_lengths = torch.bincount(values, minlength=self.num_partitions)
        assert ivf_lengths.size(0) == self.num_partitions

        print_memory_stats(f'RANK:{self.rank}')

        # Transforms centroid->embedding ivf to centroid->passage ivf
        _, _ = optimize_ivf(ivf, ivf_lengths, self.config.index_path_)


    def _update_metadata(self):
        config = self.config
        self.metadata_path = os.path.join(config.index_path_, 'metadata.json')
        Run().print("#> Saving the indexing metadata to", self.metadata_path, "..")

        with open(self.metadata_path, 'w') as f:
            d = {'config': config.export()}
            d['num_chunks'] = self.num_chunks
            d['num_partitions'] = self.num_partitions
            d['num_embeddings'] = self.num_embeddings
            d['avg_doclen'] = self.num_embeddings / len(self.collection)

            f.write(ujson.dumps(d, indent=4) + '\n')


def compute_faiss_kmeans(dim, num_partitions, kmeans_niters, shared_lists, return_value_queue=None):
    use_gpu = torch.cuda.is_available()
    kmeans = faiss.Kmeans(dim, num_partitions, niter=kmeans_niters, gpu=use_gpu, verbose=True, seed=123)

    sample = shared_lists[0][0]
    sample = sample.float().numpy()

    kmeans.train(sample)

    centroids = torch.from_numpy(kmeans.centroids)

    print_memory_stats(f'RANK:0*')

    if return_value_queue is not None:
        return_value_queue.put(centroids)

    return centroids


# hack the multiprocessing here
# it really should be tagged along with the infrastructure
def build_local_ivf(index_path, working_dir, num_partitions, info):
    pid_start, local_doclen, chunk_idx = info

    output_fn = os.path.join(working_dir, f"localivf.0.{chunk_idx}.pkl")
    if os.path.exists(output_fn):
        Run().print_main(f"#> Found local ivf for chunk {chunk_idx}, skip")
        return 

    local_pid = 0
    passage_offset_end = local_doclen[0]
    cluster2pid = [ set() for _ in range(num_partitions) ]

    chunk_codes = ResidualCodec.Embeddings.load_codes(index_path, chunk_idx)
    for i, clusterid in enumerate(chunk_codes):
        if i >= passage_offset_end:
            local_pid += 1
            passage_offset_end += local_doclen[local_pid]
            
        cluster2pid[clusterid].add(int(pid_start + local_pid))

    # save to disk to avoid slow IPC
    cluster2pid = { cid: pids for cid, pids in enumerate(cluster2pid) if len(pids) > 0 }
    
    pickle.dump(cluster2pid, open(output_fn, 'wb'))
    Run().print_main(f"#> Chunk {chunk_idx} done")

    # return cluster2pid

def _balance_file_size(fns):
    fns = sorted(fns, key=lambda f: Path(f).stat().st_size)
    balanced = list(more_itertools.interleave_longest(fns[0::2], fns[1::2][::-1]))
    assert len(fns) == len(balanced)
    return balanced


def _load_file(fn):
    if str(fn).endswith('.pt'):
        return torch.load(fn)
    elif str(fn).endswith('.pkl'):
        return pickle.load(open(fn, 'rb'))

def merge_ivf(working_dir, current_it, info):
    running_id, fns = info

    cluster2pid = { cid: list(pids) for cid, pids in _load_file(fns[0]).items() }
    for fn in fns[1:]:
        for idx, pids in _load_file(fn).items():
            if idx not in cluster2pid:
                cluster2pid[idx] = []
            cluster2pid[idx] += list(pids)
    
    pickle.dump(cluster2pid, (Path(working_dir) / f"localivf.{current_it}.{running_id}.pkl").open('wb'))

    if current_it > 1:
        for fn in fns:
            Path(fn).unlink()

    Run().print_main(f"#> Merging id #{running_id} of iteration {current_it} done")

"""
TODOs:

1. Notice we're using self.config.bsize.

2. Consider saving/using heldout_avg_residual as a vector --- that is, using 128 averages!

3. Consider the operations with .cuda() tensors. Are all of them good for OOM?
"""



import os
import torch
from pathlib import Path

# from colbert.utils.runs import Run
from colbert.infra import ColBERTConfig
from colbert.utils.utils import print_message, save_checkpoint
from colbert.parameters import SAVED_CHECKPOINTS
# from colbert.infra.run import Run

def find_last_checkpoint(path: str):
    return str(Path(path) / "colbert")

def print_progress(scores):
    positive_avg, negative_avg = round(scores[:, 0].mean().item(), 2), round(scores[:, 1].mean().item(), 2)
    print("#>>>   ", positive_avg, negative_avg, '\t\t|\t\t', positive_avg - negative_avg)


def manage_checkpoints(config: ColBERTConfig, colbert, optimizer, batch_idx, savepath=None, consumed_all_triples=False):
    # arguments = dict(args)

    # TODO: Call provenance() on the values that support it??

    # checkpoints_path = savepath or os.path.join(Run().path_, 'checkpoints')
    checkpoints_path = savepath or config.checkpoint_path_
    name = None

    try:
        save = colbert.save
    except:
        save = colbert.module.save

    if not os.path.exists(checkpoints_path):
        os.makedirs(checkpoints_path)
    
    path_save = None

    if consumed_all_triples or (batch_idx % 2000 == 0):
        # name = os.path.join(path, "colbert.dnn")
        # save_checkpoint(name, 0, batch_idx, colbert, optimizer, arguments)
        path_save = os.path.join(checkpoints_path, "colbert")

    if batch_idx in SAVED_CHECKPOINTS:
        # name = os.path.join(path, "colbert-{}.dnn".format(batch_idx))
        # save_checkpoint(name, 0, batch_idx, colbert, optimizer, arguments)
        path_save = os.path.join(checkpoints_path, f"colbert-{batch_idx}")

    if path_save:
        print(f"#> Saving a checkpoint to {path_save} ..")

        checkpoint = {}
        checkpoint['batch'] = batch_idx
        checkpoint['epoch'] = 0
        # checkpoint['model_state_dict'] = model.state_dict()
        checkpoint['optimizer_state_dict'] = optimizer.state_dict()
        # checkpoint['config'] = config

        save(path_save)
        torch.save(checkpoint, os.path.join(path_save, "misc.bin"))

    return path_save

def load_checkpoint_misc(config: ColBERTConfig):
    if config.checkpoint.endswith('.dnn'):
        # lagacy ColBERT v1 checkpoints
        raw_checkpoint = torch.load(config.checkpoint, map_location='cpu')
        if config.fix_broken_optimizer_state and min(raw_checkpoint['optimizer_state_dict']['state'].keys()) != 0:
            print_message(f"Found potentially broken optimizer state -- trying to fix it")
            print_message(f"This fix is specific for ColBERT-X with XLMR-large checkpoints created at the early stage.")

            new_state = {
                i: raw_checkpoint['optimizer_state_dict']['state'][390+i]
                for i in range(1, 390)
            }
            new_state[0] = raw_checkpoint['optimizer_state_dict']['state'][782]
            raw_checkpoint['optimizer_state_dict']['state'] = new_state
            raw_checkpoint['optimizer_state_dict']['param_groups'][0]['params'] = list(range(0, 392))

        return raw_checkpoint['batch'], raw_checkpoint['optimizer_state_dict']

    misc_fn = Path(config.checkpoint) / "misc.bin"
    assert misc_fn.exists(), f"{misc_fn} does not exists. Can't resume."
    checkpoint_misc = torch.load(misc_fn, map_location=torch.device('cpu'))
    return checkpoint_misc['batch'], checkpoint_misc['optimizer_state_dict']


import os
import ujson
from random import shuffle
from tqdm.auto import tqdm

from functools import partial
from colbert.infra.config.config import ColBERTConfig
from colbert.utils.utils import print_message, zipstar
from colbert.modeling.tokenization import QueryTokenizer, DocTokenizer, tensorize_triples
from colbert.evaluation.loaders import load_collection

from colbert.data.collection import Collection
from colbert.data.queries import Queries
from colbert.data.examples import Examples

from colbert.utils.runs import Run

from scipy.stats import beta
import numpy as np
def sampling_on_beta(items, nway: int, step: int, max_step: int, max_beta: float):
    samp_dist: np.ndarray = beta(1.0, 1.+step/max_step*(max_beta-1.)).pdf(np.arange(len(items))/len(items))
    samp_dist = samp_dist / samp_dist.sum()
    idx = np.random.choice(len(items), size=nway, replace=False, p=samp_dist)
    return [ items[i] for i in idx ]


class LazyBatcher():
    def __init__(self, config: ColBERTConfig, triples, queries=None, collection=None, rank=0, nranks=1):
        self.config = config

        self.bsize, self.accumsteps = config.bsize, config.accumsteps
        self.nway = config.nway
        self.shuffle_pasages = config.shuffle_passages
        self.sampling_max_beta = config.sampling_max_beta
        self.over_one_epoch = config.over_one_epoch

        self.query_tokenizer = QueryTokenizer(config)
        self.doc_tokenizer = DocTokenizer(config)
        self.tensorize_triples = partial(tensorize_triples, self.query_tokenizer, self.doc_tokenizer)
        self.position = 0

        if self.shuffle_pasages:
            self.triples = Examples.cast(triples, nway=100000).tolist(rank, nranks)
        else:
            self.triples = Examples.cast(triples, nway=self.nway).tolist(rank, nranks)

        # print(len(self.triples), self.triples[:10])
        self.queries = queries and Queries.cast(queries)
        self.collection = collection and Collection.cast(collection)

        self.skipping = False # only used when skipping batches

    def __iter__(self):
        return self

    def __len__(self):
        return len(self.triples)

    def __getitem__(self, idx: int):
        # TODO: Eugene [distillation] do some curriculum for this sampling
        # Curriculum Learning for Dense Retrieval Distillation: Zeng, Zamani, Vinay
        query, *passages = self.triples[ idx % len(self.triples) ]
        if self.shuffle_pasages:
            if self.sampling_max_beta > 1:
                return [ 
                    query, 
                    *sampling_on_beta(
                        passages, self.nway, idx//self.bsize, 
                        self.config.maxsteps, self.sampling_max_beta
                    ) 
                ]
            else:
                shuffle(passages)
        return [ query, *passages ]

    def __next__(self):
        # offset, endpos = self.position, min(self.position + self.bsize, len(self.triples))
        offset, endpos = self.position, self.position + self.bsize
        self.position = endpos

        if not self.over_one_epoch and offset + self.bsize > len(self.triples):
            raise StopIteration

        all_queries, all_passages, all_scores = [], [], []

        for position in range(offset, endpos):
            query, *pids = self[position]
            if self.skipping:
                continue

            pids = pids[:self.nway]

            query = self.queries[query] if self.queries else query

            try:
                pids, scores = zipstar(pids)
            except:
                scores = []

            passages = self.collection[pids] if self.collection else pids

            all_queries.append(query)
            all_passages.extend(passages)
            all_scores.extend(scores)
        
        if self.skipping:
            return 

        assert len(all_scores) in [0, len(all_passages)], len(all_scores)

        # print(all_queries)

        return self.collate(all_queries, all_passages, all_scores)

    def collate(self, queries, passages, scores):        
        assert len(queries) == self.bsize
        assert len(passages) == self.nway * self.bsize

        return self.tensorize_triples(queries, passages, scores, self.bsize // self.accumsteps, self.nway)

    def skip_to_batch(self, batch_idx):
        print(f'>>> Skipping to batch #{batch_idx} for training.')
        self.skipping = True
        for _ in tqdm(range(batch_idx), desc='skipping', dynamic_ncols=True):
            next(self)
        self.skipping = False # !!! important to turn it back on
        print(f">>> Done skipping.")

class MultiLangBatcher(LazyBatcher):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert self.collection is not None
        
        self.nlang = len(next(iter(self.collection)))

    def collate(self, queries, passages, scores):
        # each passage has self.nlang variants --> flatten them 
        assert all( len(p) == self.nlang for p in passages )
        
        queries = queries*self.nlang
        passages = sum(map(list, zip(*passages)), [])

        if len(scores) != 0:
            scores = scores*self.nlang

        assert len(queries) == self.bsize * self.nlang
        assert len(passages) == self.nway * self.bsize * self.nlang

        # print(len(queries), len(passages), len(scores))
        # print(queries, passages, scores)
        # print(passages)

        return self.tensorize_triples(queries, passages, scores, self.bsize * self.nlang // self.accumsteps, self.nway)


import os
import random
import time
import torch
import torch.nn as nn
import numpy as np

from transformers import AdamW
from colbert.infra import ColBERTConfig, Run
from colbert.utils.amp import MixedPrecisionManager

from colbert.training.lazy_batcher import LazyBatcher
from colbert.parameters import DEVICE

from colbert.modeling.colbert import ColBERT
from colbert.utils.utils import print_message
from colbert.training.utils import print_progress, manage_checkpoints, find_last_checkpoint, load_checkpoint_misc

from functools import partial
from colbert.modeling.tokenization.utils import legacy_tensorize_triples

def train(config: ColBERTConfig, triples, queries=None, collection=None):
    # lagacy training code is only compatible with
    assert config.nway == 2

    if config.resume:
        config.checkpoint = config.checkpoint or find_last_checkpoint(config.checkpoint_path_)
    else: 
        config.checkpoint = config.checkpoint or config.model_name



    random.seed(12345)
    np.random.seed(12345)
    torch.manual_seed(12345)
    
    torch.cuda.manual_seed_all(12345)

    
    assert config.bsize % config.nranks == 0, (config.bsize, config.nranks)
    assert config.accumsteps == 1
    config.bsize = config.bsize // config.nranks

    print("Using config.bsize =", config.bsize, "(per process) and config.accumsteps =", config.accumsteps)

    
    reader = LazyBatcher(config, triples, queries, collection, (0 if config.rank == -1 else config.rank), config.nranks)
    # reader.tensorize_triples = partial(legacy_tensorize_triples, reader.query_tokenizer, reader.doc_tokenizer)

    # if config.rank not in [-1, 0]:
    #     torch.distributed.barrier()

    colbert = ColBERT(name=config.checkpoint, colbert_config=config)

    
    # TODO: add load optimizer back     
    # if config.checkpoint is not None:
    #     # assert config.resume_optimizer is False, "TODO: This would mean reload optimizer too."
    #     if not config.resume_optimizer:
    #         print_message(f"#> Starting from checkpoint {config.checkpoint} -- but NOT the optimizer!")
    #     else:
    #         print_message(f"#> Starting from checkpoint {config.checkpoint}")

    #     checkpoint = torch.load(config.checkpoint, map_location='cpu')

    #     try:
    #         colbert.load_state_dict(checkpoint['model_state_dict'])
    #     except:
    #         print_message("[WARNING] Loading checkpoint with strict=False")
    #         colbert.load_state_dict(checkpoint['model_state_dict'], strict=False)

    # if config.rank == 0:
    #     torch.distributed.barrier()

    colbert = colbert.to(DEVICE)
    colbert.train()


    colbert = torch.nn.parallel.DistributedDataParallel(colbert, device_ids=[config.rank],
                                                        output_device=config.rank,
                                                        find_unused_parameters=True)

    # Attempt to load optimizer
    optimizer = AdamW(filter(lambda p: p.requires_grad, colbert.parameters()), lr=config.lr, eps=1e-8)
    # if config.resume_optimizer:
    #     assert config.checkpoint is not None
    #     print_message("#> Loading the optimizer")
    #     optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
    optimizer.zero_grad()

    amp = MixedPrecisionManager(config.amp)
    criterion = nn.CrossEntropyLoss()
    labels = torch.zeros(config.bsize, dtype=torch.long, device=DEVICE)

    start_time = time.time()
    train_loss = 0.0

    start_batch_idx = 0

    # if config.resume:
    #     assert config.checkpoint is not None
    #     start_batch_idx = checkpoint['batch']

    #     reader.skip_to_batch(start_batch_idx, checkpoint['arguments']['bsize'])

    for batch_idx, BatchSteps in zip(range(start_batch_idx, config.maxsteps), reader):
        this_batch_loss = 0.0

        for queries, passages, _ in BatchSteps:
            with amp.context():
                # queries = ( torch.concat((queries[0], queries[0])), torch.concat((queries[1], queries[1])) )
                # passages = ( torch.concat((passages[0][0::2], passages[0][1::2])), torch.concat((passages[1][0::2], passages[1][1::2])) )

                # queries = (
                #     queries[0].repeat_interleave(2, dim=0).contiguous(),
                #     queries[1].repeat_interleave(2, dim=0).contiguous()
                # )

                scores = colbert(queries, passages)
                # print(scores.tolist())
                
                scores = scores.view(-1, config.nway)

                # .view(-1, config.nway)
                
                # .view(2, -1).permute(1, 0)

                loss = criterion(scores, labels[:scores.size(0)])
                loss = loss / config.accumsteps

            if config.rank < 1:
                print_progress(scores)

            amp.backward(loss)

            train_loss += loss.item()
            this_batch_loss += loss.item()

        amp.step(colbert, optimizer)

        if config.rank < 1:
            avg_loss = train_loss / (batch_idx+1)

            # num_examples_seen = (batch_idx - start_batch_idx) * config.bsize * config.nranks
            # elapsed = float(time.time() - start_time)

            # log_to_mlflow = (batch_idx % 20 == 0)
            # Run.log_metric('train/avg_loss', avg_loss, step=batch_idx, log_to_mlflow=log_to_mlflow)
            # Run.log_metric('train/batch_loss', this_batch_loss, step=batch_idx, log_to_mlflow=log_to_mlflow)
            # Run.log_metric('train/examples', num_examples_seen, step=batch_idx, log_to_mlflow=log_to_mlflow)
            # Run.log_metric('train/throughput', num_examples_seen / elapsed, step=batch_idx, log_to_mlflow=log_to_mlflow)

            print_message(batch_idx, avg_loss)
            manage_checkpoints(config, colbert, optimizer, batch_idx+1)
            
    manage_checkpoints(config, colbert, optimizer, batch_idx+1, consumed_all_triples=True)


import os
import argparse

from colbert.infra import Run, RunConfig, ColBERTConfig
from colbert import Indexer

from colbert.scripts.collection_utils import load_irds_or_local

def create_plaid_index(args):

    if args.n_gpus is None:
        import torch
        n_gpus = torch.cuda.device_count()
    else:
        n_gpus = args.n_gpus

    # Provide v1 checkpoint configurations
    config = ColBERTConfig(
        doc_maxlen=args.doc_maxlen, nbits=args.nbits,
        max_sampled_pid=args.max_sampled_pid, max_num_partitions=args.max_num_partitions
    )

    coll_file = str(os.path.join(args.coll_dir, 'collection_passages.tsv')) if os.path.isdir(args.coll_dir) else args.coll_dir
    collection = load_irds_or_local(
        args.dataset_name if args.coll_dir is None else coll_file,
        component='docs', use_offsetmap=args.lazy_collection_loader
    )
    dataset_name = args.dataset_name.replace('/', '.')

    if args.base_model is not None:
        config.configure(model_name=args.base_model)
    if args.checkpoint.endswith('.dnn'):
        # specifically for our ColBERT-X checkpoints
        config.configure(force_resize_embeddings=True, mask_punctuation=False)

    index_name = args.index_name or f'{dataset_name}.{args.nbits}bits'
    config.configure(
        index_name=index_name, 
        use_lagacy_build_ivf=args.use_lagacy_build_ivf,
        reuse_centroids_from=args.reuse_centroids_from
    )

    run_config = RunConfig(
        nranks=n_gpus, root=args.root, 
        ivf_num_processes=args.ivf_num_processes, 
        ivf_use_tempdir=args.ivf_use_tempdir, 
        ivf_merging_ways=args.ivf_merging_ways,
        experiment=args.experiment, 
        index_root=args.index_root
    )

    with Run().context(run_config):  # nranks specifies the number of GPUs to use.
        indexer = Indexer(checkpoint=args.checkpoint, config=config)

        if args.step == 'prepare':
            indexer.prepare(name=index_name, collection=collection, overwrite='resume')
        elif args.step == 'encode':
            indexer.encode(name=index_name, collection=collection)
        elif args.step == 'finalize':
            indexer.finalize(name=index_name, collection=collection)
            print(indexer.get_index()) # You can get the absolute path of the index, if needed.
            print("Index created")


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--coll_dir', help="directory containing collection_passages.tsv file", default=None)
    parser.add_argument('--index_name', help="name of the index, default {dataset_name}.{nbits}bits", type=str, default=None)
    parser.add_argument('--dataset_name', help='name of the dataset', required=True)

    parser.add_argument('--step', help="step of indexing", choices=['prepare', 'encode', 'finalize'], required=True)
    parser.add_argument('--nbits', help='number of bits used to compress index', type=int, default=2)
    parser.add_argument('--doc_maxlen', help="the number used to truncate passages", type=int, default=300)

    parser.add_argument('--checkpoint', help="path to the model checkpoint", type=str, required=True)
    parser.add_argument('--base_model', help="base model of the v1 checkpoint", type=str, default=None)

    parser.add_argument('--index_root', help="directory where index is written", type=str, default=None)
    parser.add_argument('--experiment', help="name of the experiment", type=str, default="default")
    parser.add_argument('--root', type=str, default=None)
    parser.add_argument('--n_gpus', help="number of GPU, default using all available ones", type=int, default=None)

    parser.add_argument('--max_num_partitions', help="max number of partitions/centroid used in indexing", type=int, default=-1)
    parser.add_argument('--max_sampled_pid', help="max number of sampled tokens for training KMeans", type=int, default=-1)

    parser.add_argument('--lazy_collection_loader', action='store_true', default=False)
    parser.add_argument('--ivf_num_processes', type=int, default=20)
    parser.add_argument('--ivf_use_tempdir', action='store_true', default=False)
    parser.add_argument('--ivf_merging_ways', type=int, default=2)

    parser.add_argument('--use_lagacy_build_ivf', action='store_true', default=False)    
    parser.add_argument('--use_offset_map', action='store_true', default=False)

    parser.add_argument('--reuse_centroids_from', type=str, default=None)

    create_plaid_index(parser.parse_args())


import argparse
from colbert.infra import Run, RunConfig, ColBERTConfig
from colbert import Trainer
from colbert.utils.utils import print_message

from colbert.scripts.collection_utils import load_irds_or_local

def main(args):
    bsize = args.per_device_batch_size*args.n_gpus
    exp_name = f"{args.run_tag}/{bsize}bat.{args.nway}way{'-ib' if args.ib_negatives else ''}"

    if args.use_local_full_triples:
        queries = None
        collection = None
    elif args.training_queries is not None:
        queries = load_irds_or_local(args.training_queries, 'queries')
        collection = load_irds_or_local(args.training_collection, 'docs', mixing=args.training_collection_mixing, use_offsetmap=args.use_offsetmap)
    else:
        queries = load_irds_or_local(args.training_irds_id, 'queries')
        collection = load_irds_or_local(args.training_irds_id, 'docs', use_offsetmap=args.use_offsetmap)

    triples = load_irds_or_local(args.training_triples, 'docpairs')

    # parse other_args
    other_args = {}
    for k,v in map(lambda x: x.split('='), args.other_args):
        try: 
            other_args[k] = eval(v)
        except:
            other_args[k] = v
        
    with Run().context(RunConfig(nranks=args.n_gpus, root=args.root, experiment=args.experiment, name=exp_name)):
        trainer = Trainer(
            triples=triples, queries=queries, collection=collection,
            config=ColBERTConfig(
                bsize=bsize, 
                checkpoint=args.checkpoint,
                maxsteps=args.maxsteps, lr=args.learning_rate, nway=args.nway, 
                accumsteps=args.accumsteps, 
                model_name=args.model_name, 
                similarity=args.similarity,
                force_resize_embeddings=True,
                doc_maxlen=args.doc_maxlen, 
                use_ib_negatives=args.ib_negatives,
                ignore_scores=(args.kd_loss == 'None'),
                kd_loss=args.kd_loss, 
                shuffle_passages=not args.only_top, 
                sampling_max_beta=args.sampling_max_beta,
                over_one_epoch=True,
                resume=args.resume,
                resume_optimizer=args.resume_optimizer,
                fix_broken_optimizer_state=args.fix_broken_optimizer_state,
                multilang=(args.training_collection_mixing == 'all'),
                **other_args
            )
        )

        trainer.train()

        print_message(f"Best model: {trainer.best_checkpoint_path()}")
        
        
if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--model_name', type=str, default='xlm-roberta-large')
    parser.add_argument('--checkpoint', type=str, default=None)
    
    parser.add_argument('--use_local_full_triples', action='store_true', default=False)
    parser.add_argument('--training_triples', type=str, required=True) # can be a irds id or path
    parser.add_argument('--training_irds_id', type=str, default='beir/msmarco/train')
    parser.add_argument('--training_queries', type=str, default=None)
    parser.add_argument('--training_collection', type=str, nargs='+', default=None)

    parser.add_argument('--doc_maxlen', type=int, default=220)
    parser.add_argument('--use_offsetmap', action='store_true', default=False)

    parser.add_argument('--training_collection_mixing', choices=['entries', 'passages', 'elements', 'all', 'round-robin'], default='elements')
    parser.add_argument('--similarity', choices=['l2', 'cosine'], default='cosine')

    parser.add_argument('--learning_rate', type=float, default=5e-6)
    parser.add_argument('--nway', type=int, default=8)
    parser.add_argument('--per_device_batch_size', type=int, default=8)
    parser.add_argument('--maxsteps', type=int, default=200_000)
    parser.add_argument('--ib_negatives', action='store_true', default=False)
    parser.add_argument('--accumsteps', type=int, default=1)

    parser.add_argument('--kd_loss', choices=['KLD', 'MSE', 'None'], default='KLD')

    parser.add_argument('--only_top', action='store_true', default=False)
    parser.add_argument('--sampling_max_beta', type=float, default=1.0)

    parser.add_argument('--experiment', help="name of the experiment for locating the index", type=str, default=None)
    parser.add_argument('--root', type=str, default=None)
    parser.add_argument('--run_tag', type=str, default=None)
    parser.add_argument('--n_gpus', help="number of gpus to use, default all visible ones in CUDA_VISIBLE_DEVICES", type=int, default=None)

    parser.add_argument('--resume', action='store_true', default=False)
    parser.add_argument('--resume_optimizer', action='store_true', default=False)
    parser.add_argument('--fix_broken_optimizer_state', action='store_true', default=False)
    parser.add_argument('--other_args', type=str, nargs='+', default=[])

    args = parser.parse_args()

    if args.n_gpus is None:
        import torch
        args.n_gpus = torch.cuda.device_count()

    # make sure when running distillation, the triple file is a jsonl file
    if args.kd_loss != 'None':
        assert args.training_triples.endswith('.jsonl')

    if args.training_queries is not None or args.training_collection is not None:
        assert args.training_queries is not None
        assert args.training_collection is not None
        print_message(f"Use separate irds for training queries and documents.")

    if args.training_collection_mixing == 'passages':
        args.training_collection_mixing = 'elements'
    elif args.training_collection_mixing == 'round-robin':
        args.training_collection_mixing = 'all'

    args.run_tag = args.run_tag or f"{args.learning_rate}"
    args.experiment = args.experiment or f"plaid_{args.model_name.split('/')[-1]}"

    main(args)


import setuptools

with open('README.md', 'r') as f:
    long_description = f.read()

setuptools.setup(
    name='PLAID-X',
    version='0.3.2.post2',
    author='Eugene Yang',
    author_email='eugene.yang@jhu.edu',
    description="Efficient and Effective Passage Search via Contextualized Late Interaction over BERT and XLM-RoBERTa",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/hltcoe/ColBERT-X/tree/plaid-x',
    packages=setuptools.find_packages(),
    install_requires=open('requirements.txt').read().split("\n"),
    extras_require={
        'gpu': ['faiss-gpu']
    },
    include_package_data=True,
    python_requires='>=3.8',
)



import os
import time

import torch.multiprocessing as mp

from colbert.infra.run import Run
from colbert.infra.config import ColBERTConfig, RunConfig
from colbert.infra.launcher import Launcher

from colbert.utils.utils import create_directory, print_message

from colbert.indexing.collection_indexer import reuse_prepare, sample, kmeans, encode, finalize

class Indexer:
    def __init__(self, checkpoint, config=None):
        """
           Use Run().context() to choose the run's configuration. They are NOT extracted from `config`.
        """

        self.index_path = None
        self.checkpoint = checkpoint
        self.checkpoint_config = ColBERTConfig.load_from_checkpoint(checkpoint)

        self.config = ColBERTConfig.from_existing(self.checkpoint_config, config, Run().config)
        self.configure(checkpoint=checkpoint)

    def configure(self, **kw_args):
        self.config.configure(**kw_args)

    def get_index(self):
        return self.index_path

    def erase(self):
        assert self.index_path is not None
        directory = self.index_path
        deleted = []

        for filename in sorted(os.listdir(directory)):
            filename = os.path.join(directory, filename)

            delete = filename.endswith(".json")
            delete = delete and ('metadata' in filename or 'doclen' in filename or 'plan' in filename)
            delete = delete or filename.endswith(".pt")
            
            if delete:
                deleted.append(filename)
        
        if len(deleted):
            print_message(f"#> Will delete {len(deleted)} files already at {directory} in 20 seconds...")
            time.sleep(20)

            for filename in deleted:
                os.remove(filename)

        return deleted

    def prepare(self, name, collection, overwrite=False, no_kmeans=False):
        assert overwrite in [True, False, 'reuse', 'resume']

        self.configure(collection=collection, index_name=name, resume=overwrite=='resume')
        self.configure(partitions=None)

        self.index_path = self.config.index_path_
        index_does_not_exist = (not os.path.exists(self.config.index_path_))

        assert (overwrite in [True, 'reuse', 'resume']) or index_does_not_exist, self.config.index_path_
        create_directory(self.config.index_path_)

        if overwrite is True:
            self.erase()

        if index_does_not_exist or overwrite != 'reuse':
            self.__launch(sample, collection, nospawn=self.config.gpus == 1)
            if not no_kmeans:
                self.__launch(kmeans, collection, nospawn=True)

        return self.index_path

    def encode(self, name, collection):
        self.configure(collection=collection, index_name=name, resume=True)
        self.configure(partitions=None)

        self.index_path = self.config.index_path_
            
        if self.config.reuse_centroids_from is not None: 
            plan_path = os.path.join(self.index_path, "plan.json")
            if not os.path.exists(plan_path):
                # do a light prepare step to create the plan.json file along and soft link the centriods 
                self.index_path = self.config.index_path_
                create_directory(self.config.index_path_)
                self.__launch(reuse_prepare, collection, nospawn=True)
            else: 
                print_message(f"############# Ignoring `reuse_centroids_from` since the plan in the index directory already exists -- will only try to resume indexing only")

        assert os.path.exists(self.config.index_path_), "Run first step `prepare` in advance."

        self.__launch(encode, collection, nospawn=self.config.gpus == 1)

        return self.index_path

    def finalize(self, name, collection):
        self.configure(collection=collection, index_name=name, resume=True)
        self.configure(bsize=64, partitions=None)

        self.index_path = self.config.index_path_
            
        

        self.__launch(finalize, collection, nospawn=True)

        return self.index_path


    def __launch(self, callee, collection, nospawn=False):
        manager = mp.Manager()
        shared_lists = [manager.list() for _ in range(self.config.nranks)]
        shared_queues = [manager.Queue(maxsize=1) for _ in range(self.config.nranks)]

        if not nospawn:
            launcher = Launcher(callee)
            launcher.launch(self.config, collection, shared_lists, shared_queues)
        else:
            new_config = type(self.config).from_existing(self.config, RunConfig.from_existing(Run().config))
            callee(new_config, collection, shared_lists, shared_queues)
